import numpy as np, planer
from . import cellpose
from .cellpose import load_model, count_flow, get_flow, flow2msk
from .render import msk2edge, rgb_mask, draw_edge, flow2hsv, show
from .demo import mini_test, tile_test
from urllib.request import urlretrieve, urlopen

from glob import glob
import os, os.path as osp
from tqdm import tqdm

def engine(core, nimg):
    global _asnumpy
    cellpose.np, cellpose.ndimg = core, nimg
    _asnumpy = planer.core(core)
    print('\nuser switch engine:', core.__name__)

def asnumpy(arr): return _asnumpy.cpu(arr)

try:
    import cupy
    from cupyx.scipy import ndimage as ndimg
    engine(cupy, ndimg)
    print('using cupy engine, gpu powered!')
except:
    import numpy as np
    import scipy.ndimage as ndimg
    engine(np, ndimg)
    print('using numpy engine, install cupy would be faster.')

root = osp.abspath(osp.dirname(__file__))
if not osp.exists(root+'/models'): os.mkdir(root+'/models')

models = {}

def search_models():
    url = 'http://release.imagepy.org/cellpose-planer/README.md'
    lines = urlopen(url).read().decode('utf-8').split('\n')
    global models
    for line in lines:
        if not '[download](' in line: continue
        ss = line.split('|')
        url = ss[-2].split('(')[-1].split(')')[0]
        name = osp.split(url)[1]
        models[ss[1].strip()] = url
    fs = glob(root+'/models/*.npy')
    has = [osp.split(i)[1][:-4] for i in fs]
    for i in sorted(models):
        print(i, ':', 'installed' if i in has else '--')
    
def list_models():
    fs = glob(root+'/models/*.npy')
    return [osp.split(i)[1][:-4] for i in fs]

def progress(a, b, c, bar):
    per = int(min(100.0 * a * b / c, 100))
    bar.update(round(per)-bar.n)

def download(names=['cyto_0','cyto_1','cyto_2','cyto_3']):
    assert len(models)>0, 'please search_models first!'
    if names=='all': names = list(models.keys())
    if isinstance(names, str): names = [names]
    for name in names:
        print('download %s from %s'%(name, models[name]))
        bar = tqdm(total=100)
        urlretrieve(models[name], root+'/models/'+name+'.npy',
            lambda a,b,c,bar=bar: progress(a,b,c,bar))
        urlretrieve(models[name][:-3]+'json',
                    root+'/models/'+name+'.json')

def test():
    from time import time
    from skimage.data import coins
    img = coins()
    x = img.astype(np.float32)/255
    net = load_model('cyto_0')
    start = time()
    flow, prob, style = get_flow(net, x)
    print('\tnet time:', time()-start)
    start = time()
    lab = flow2msk(flow, prob, level=0.2)
    print('\tflow time:', time()-start)
    flow = asnumpy(flow)
    prob = asnumpy(prob)
    lab = asnumpy(lab)
    show(img, flow, prob, lab)



import random, numpy as np

def flow2hsv(flow):
    v = np.linalg.norm(flow, axis=-1)
    h = np.arccos(flow[:,:,0]/v)
    h *= np.sign(flow[:,:,1])/(np.pi*2)
    h += 0.5; v /= v.max()
    
    s = np.ones_like(v)             
    a = np.floor(h * 6)
    b = h * 6; b -= a
    p = np.zeros_like(v)
    t = v * b; q = v - t
    
    buf = np.stack((v,t,p,q), -1).ravel()
    buf *= 255; buf = buf.astype(np.uint8)
    idx = np.array([
        [0,1,3],[3,0,2],[2,0,1], [2,3,0],[1,2,0],[0,2,3]])
    idx = idx[a.ravel().astype(np.uint8) % 6]
    idx += np.arange(v.size)[:,None] * 4
    return buf[idx].reshape(v.shape+(3,))
   
def msk2edge(lab):
   msk = np.zeros(lab.shape, dtype=np.bool)
   mskr = lab[1:] != lab[:-1]
   mskc = lab[:,1:] != lab[:,:-1]
   msk[1:] |= mskr; msk[:-1] |= mskr
   msk[:,1:] |= mskc; msk[:,:-1] |= mskc
   return msk

def draw_edge(img, lab, color=(255,0,0)):
    edge = msk2edge(lab)
    lut = np.array([[0,0,0],color], dtype=np.uint8)
    rgb = lut[edge.view(np.uint8)]
    img = img.reshape((img.shape+(1,))[:3])
    return np.maximum(img, rgb, out=rgb)

def connect_graph(img):
    pair1 = np.concatenate((img[:-1,:,None], img[1:,:,None]), -1)
    pair2 = np.concatenate((img[:,:-1,None], img[:,1:,None]), -1)
    pair = np.vstack((pair1.reshape(-1,2), pair2.reshape(-1,2)))
    pair = pair[pair[:,0]!=pair[:,1]];pair = pair[pair.min(1)>0]
    idx = np.unique(np.sort(pair), axis=0) if len(pair)>0 else []
    dic = {}
    for i in np.unique(img): dic[i] = []
    for i,j in idx:
        dic[i].append(j)
        dic[j].append(i)
    del dic[0]
    return dic

def node_render(conmap, n=5, rand=10, shuffle=True):
   nodes = list(conmap.keys())
   colors = dict(zip(nodes, [0]*len(nodes)))
   counter = dict(zip(nodes, [0]*len(nodes)))
   if shuffle: random.shuffle(nodes)
   while len(nodes)>0:
     k = nodes.pop(0)
     counter[k] += 1
     hist = [1e4] + [0] * n
     for p in conmap[k]:
         hist[colors[p]] += 1
     if min(hist)==0:
         cand = [i for i in range(n+1) if not hist[i]]
         colors[k] = cand[random.randint(0, len(cand)-1)]
         counter[k] = 0
         continue
     hist[colors[k]] = 1e4
     minc = hist.index(min(hist))
     if counter[k]==rand:
         counter[k] = 0
         minc = random.randint(1,n)
     colors[k] = minc
     for p in conmap[k]:
         if colors[p] == minc:
             nodes.append(p)
   lut = np.zeros(len(colors)+1, dtype=np.uint8)
   for c in colors: lut[c] = colors[c]
   return lut

def rgb_mask(img, lab):
   cmap = np.array([(0,0,0),(255,0,0),(0,255,0),
      (0,0,255),(255,255,0),(255,0,255)], dtype=np.uint8)
   idx = connect_graph(lab)
   lut = node_render(idx)
   rgb = cmap[lut][lab]
   img = img.reshape((img.shape+(1,))[:3])
   return np.maximum(img, rgb, out=rgb)

def show(img, flowpb, lab):
    import matplotlib.pyplot as plt
    plt.subplot(221).imshow(img)
    plt.subplot(222).imshow(draw_edge(img, lab))
    plt.subplot(223).imshow(flow2hsv(flowpb[:,:,:2]))
    plt.subplot(224).imshow(rgb_mask(img,lab))
    plt.show()



from planer import read_net, resize
import numpy as np
import scipy.ndimage as ndimg
import random, math, itertools
from tqdm import tqdm

import os.path as osp
root = osp.abspath(osp.dirname(__file__))

def progress(n, i, bar=[None]):
   if bar[0] is None:
      bar[0] = tqdm()
   bar[0].total = n
   bar[0].update(1)
   if n==i: bar[0] = None

def load_model(names='cyto_0'):
   if isinstance(names, str): return read_net(root+'/models/'+names)
   return [read_net(root+'/models/'+i) for i in names]

def count_flow(nets, img, cn=[0,0], size=512, work=1):
   if not isinstance(nets, list): nets = [nets]
   img = np.asarray(img)[None, cn, :, :]
   h, w = img.shape[-2:]
   if not(h==w==size): img = resize(img, (size,size))
   y = np.zeros((1,3)+img.shape[2:], img.dtype)
   style = np.zeros((1,256), img.dtype)
   def one(net, img):
      i, s = net(img)
      y[:] += i; style[:] += s
   if work>1 and len(nets)>1:
      from concurrent.futures import ThreadPoolExecutor
      pool = ThreadPoolExecutor(max_workers=work, thread_name_prefix="flow")
      for net in nets: pool.submit(one, net, img)
      pool.shutdown(wait=True)
   else:
      for net in nets: one(net, img)
   if len(nets)>0:
      y /= len(nets); style /= len(nets)
   if not(h==w==size): y = resize(y, (h, w))
   return y[0].transpose(1,2,0), style

def make_slice(l, w, mar):
    r = np.linspace(w//2, l-w//2, math.ceil((l-mar)/(w-mar))).astype(int)
    return [slice(i-w//2, i+w//2) for i in r.tolist()]

def grid_slice(H, W, size, mar):
    a, b = make_slice(H, size, mar), make_slice(W, size, mar)
    return list(itertools.product(a, b))

def get_flow(nets, img, cn=[0,0], sample=1, size=512, tile=True, work=1, callback=progress):
   if not isinstance(nets, list): nets = [nets]
   if img.ndim==2: img = np.asarray(img[None,:,:])[cn]
   else: img = np.asarray(img.transpose(2,0,1))[cn]
   (_, H, W), k = img.shape, sample;
   h, w = (size, size) if not tile else (max(size,int(H*k)), max(int(W*k), size))
   needresize = ((k!=1 or min(H, W)<size) and tile) or (not(H==W==size) and not tile)
   simg = img if not needresize else resize(img[:,:,:], (h, w))
   rcs = grid_slice(h, w, size, size//10)
   flow = np.zeros((3, h, w), dtype=simg.dtype)
   style = np.zeros((1, 256), dtype=simg.dtype)
   count = np.zeros(simg.shape[1:], 'uint8')
   def one(sr, sc, sz, wk, s=[0]):
      flw_prb, sty = count_flow(nets, simg[:,sr,sc], slice(None), sz, wk)
      flow[:, sr, sc] += flw_prb.transpose(2,0,1)
      style[:] += sty
      count[sr, sc] += 1
      s[0] += 1
      callback(len(rcs), s[0])
   if work>1 and len(rcs)>1:
      from concurrent.futures import ThreadPoolExecutor
      pool = ThreadPoolExecutor(max_workers=work, thread_name_prefix="net")
      for i in range(len(rcs)): pool.submit(one, *rcs[i], size, 1)
      pool.shutdown(wait=True)
   else:
      for slr, slc in rcs: one(slr, slc, size, work)
   flow /= count; style /= len(rcs)
   if needresize: flow = resize(flow, (H, W))
   flow[2]*=-1; np.exp(flow[2], out=flow[2]);
   flow[2]+=1; np.divide(1, flow[2], out=flow[2])
   return flow.transpose(1,2,0), style

def estimate_volumes(arr, sigma=3):
    msk = arr > 50; 
    idx = np.arange(len(arr), dtype=np.uint32)
    idx, arr = idx[msk], arr[msk]
    for k in np.linspace(5, sigma, 5):
       std = arr.std()
       dif = np.abs(arr - arr.mean())
       msk = dif < std * k
       idx, arr = idx[msk], arr[msk]
    return arr.mean(), arr.std()

def flow2msk(flowp, level=0.5, grad=0.5, area=None, volume=None):
    flowp = np.asarray(flowp)
    shp, dim = flowp.shape[:-1], flowp.ndim - 1
    l = np.linalg.norm(flowp[:,:,:2], axis=-1)
    flow = flowp[:,:,:2]/l.reshape(shp+(1,))
    flow[(flowp[:,:,2]<level)|(l<grad)] = 0
    ss = ((slice(None),) * (dim) + ([0,-1],)) * 2
    for i in range(dim):flow[ss[dim-i:-i-2]+(i,)]=0
    sn = np.sign(flow); sn *= 0.5; flow += sn;
    dn = flow.astype(np.int32).reshape(-1, dim)
    strides = np.cumprod(np.array((1,)+shp[::-1]))
    dn = (strides[-2::-1] * dn).sum(axis=-1)
    rst = np.arange(flow.size//dim); rst += dn
    for i in range(10): rst = rst[rst]
    hist = np.bincount(rst, None, len(rst))
    hist = hist.astype(np.uint32).reshape(shp)
    lab, n = ndimg.label(hist, np.ones((3,)*dim))
    volumes = ndimg.sum(hist, lab, np.arange(n+1))
    areas = np.bincount(lab.ravel())
    mean, std = estimate_volumes(volumes, 2)
    if not volume: volume = max(mean-std*3, 50)
    if not area: area = volumes // 3
    msk = (areas<area) & (volumes>volume)
    lut = np.zeros(n+1, np.uint32)
    lut[msk] = np.arange(1, msk.sum()+1)
    return lut[lab].ravel()[rst].reshape(shp)
    return hist, lut[lab], mask
   
if __name__ == '__main__':
   a = np.random.rand(100)
   a[3] = 2
   idx = filter(a)



import sys; sys.path.append('../')
import cellpose_planer as cellpp
import numpy as np
from skimage.data import coins

def mini_test():
    print('coins test:')
    img = coins()
    x = img.astype(np.float32)/255
    net = cellpp.load_model('cyto_0')
    flowpb, style = cellpp.get_flow(net, x, size=480)
    lab = cellpp.flow2msk(flowpb, level=0.2)
    flowpb, lab = cellpp.asnumpy(flowpb), cellpp.asnumpy(lab)
    cellpp.show(img, flowpb, lab)

def tile_test():
    print('\nlarge tile test:')
    img = np.tile(np.tile(coins(), 4).T, 4).T
    img = img[:,:,None][:,:,[0,0,0]]
    x = img.astype(np.float32)/255
    net = cellpp.load_model(['cyto_0'])
    flowpb, style = cellpp.get_flow(net, x, sample=1, size=512, tile=True, work=1)
    lab = cellpp.flow2msk(cellpp.asnumpy(flowpb), area=20, volume=50)
    flowpb, lab = cellpp.asnumpy(flowpb), cellpp.asnumpy(lab)
    cellpp.show(img, flowpb, lab)
    
if __name__ == '__main__':
    mini_test()
    tile_test()
    



from setuptools import setup

descr = """Cellpose-Planer is the cellpose models on planer framework"""

if __name__ == '__main__':
    setup(name='cellpose-planer',
        version='0.14',
        url='https://github.com/Image-Py/cellpose-planer',
        description='Cellpose-Planer is the cellpose models on planer framework',
        long_description=descr,
        author='Y.Dong, YXDragon',
        author_email='yxdragon@imagepy.org',
        license='BSD 3-clause',
        packages=['cellpose_planer'],
        package_data={},
        install_requires=[
            'numpy',
            'scipy',
            'planer'
        ],
    )
